/**
 * A MuSocket is used to hold all information important to
 * a single socket connection
 * $Id: MuSocket.java 1.25 2003/01/13 05:09:33 jeffnik Exp $
 */

/* JamochaMUD, a Muck/Mud client program
 * Copyright (C) 1998-2002 Jeff Robinson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package anecho.JamochaMUD;

import anecho.gui.JMText;
import anecho.gui.ResReader;

import anecho.extranet.Socks5socket;
import anecho.extranet.MUDBufferedReader;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import java.io.DataOutputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.StringWriter;    // This is for debugging purposes
import java.io.PrintWriter;    // This is for debugging purposes

import java.net.ConnectException;
import java.net.InetAddress;
import java.net.NoRouteToHostException;
import java.net.Socket;
import java.net.UnknownHostException;

import java.util.Vector;

public class MuSocket extends Thread implements MouseListener {

    private String name, address;
    private int port;
    private long timeStamp;
    private int activity = 0;   // Keep track of new activity if we're not the activeMU
    private boolean activeStatus = true;   // Is this connection the "active" one?
    private boolean connected;        // The state of the connection
    private boolean paused;             // Has our output been paused?
    private boolean echo = false;        // Status of "IAC ECHO"
    private JMText mainText;            // The text-area belonging to this connection
    private JMConfig settings;          // A pointer to our settings
    private Socket connSock;
    private String fromMU;              // a string to hold output from the MU*
    private MUDBufferedReader inStream; // input to us from the MU*
    private DataOutputStream outStream; // for sending info to the MU*
    private Vector heldResponse;        // A vector to hold any text incase we are "paused"
    private int minimisedLineCount = 0; // the number of minimised lines
    private CHandler connHandler;

    // Telnet protocols
    private static final String DO = "\u00FD";
    private static final String IAC = "\u00a0";
    private static final String ECHO = "\u0001";
    private static final String EOR = "\u0019";       // End of record (line) XXX
    private static final String WILL = "\u00b9";
    private static final String WONT = "\u00b3";
    private static final String TERMINAL_TYPE = "\u0018";
    private static final String IAC_WILL_ECHO = IAC + WILL + ECHO;
    private static final String IAC_WONT_ECHO = IAC + WONT + ECHO;
    private static final String IAC_WILL_EOR = IAC + WILL + EOR;
    private static final String IAC_DO_EOR = IAC + DO + EOR;

    private static final String ESCAPE = "\u001B";
    private static final String B2 = "\u00B2";
    private static final String BELL = "\u0007";        // Bell sound
    // private static final String B9 = "\u00b9";  // superscript "1" character
    // private static final String FACE = "\u263A";        // smiley face character
    // private static final String FACE = "\u0001";        // smiley face character (SOH) ***ECHO!!!***
    private static final String CAN = "\u0018";     // CAN - Cancel
    private static final String EM = "\u0019";  // EM - End Of Medium XXX
    private static final String US = "\u001f";     // US - Unit separator

    private static final boolean DEBUG = false;
    /** The constructor.  Behold!! */
    public MuSocket(JMConfig settings) {
        this.settings = settings;
        // Set up a new text area
        mainText = new JMText("", 50, 80, JMText.SCROLLBARS_VERTICAL_ONLY);
        // Set the attributes of our text area
        mainText.setEditable(false);
        heldResponse = new Vector(0, 1);

        mainText.setDoubleBuffer(settings.getDoubleBuffer());

        mainText.addMouseListener(this);

        connHandler = settings.getConnectionHandler();

        setMainColours();
    }

    // Check for mouse actions

    public void mousePressed(MouseEvent e) {
        setPaused(true);
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseClicked(MouseEvent e) {
        // Determine number of mouse clicks
        int a = e.getClickCount();

        if (a >= 2) {
            // Call outside routines (plug-ins) to handle the selected text
            e.consume();
            SpoolText();

            // grab the 'selected' URL
            StringBuffer tentativeURL = new StringBuffer((mainText.getSelectedText()).trim());
            // System.out.println("Muckmain... using " + tentativeURL + " for external program.");

            // Fix this XXX
            // The frame is sent to anchor any potential dialogues
            // send the 'URL' to launch the correct program
            ExternalProgs ep = new ExternalProgs(settings.getMainWindowVariable(), settings);
            ep.LaunchProgram(settings.getMainWindowVariable(), tentativeURL);

            // As a quick fix we'll copy this to the Clipboard instead.
            java.awt.datatransfer.Clipboard c = mainText.getToolkit().getSystemClipboard();

            try {
                java.awt.datatransfer.StringSelection s = new java.awt.datatransfer.StringSelection(tentativeURL.toString());
                c.setContents(s, s);
            } catch (Exception clipOopse) {
                System.out.println("MuSocket exception copying to Clipboard." + clipOopse);
            }


            // Deselect the text
            // mainText.select(mainText.getSelectionEnd(), mainText.getSelectionEnd());
            // System.out.println("MuSocket: Deselect text?");
        } else {
            // A single click signals a pause
            // This will pause the output window,
            // the incoming lines held in queue
            setPaused(true);
            // PauseText();
        }
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }

    /** The name of our connection */
    public synchronized void setTitle(String name) {
        this.name = name;
        setFrameTitles();
    }

    /** getTitle does a small bit of book-keeping, returning a title based
     * on whether we're connected to the MU* or not.  If we're not connected
     * then we'll return &quot;JamochaMUD" instead of the name
     */
    public synchronized String getTitle() {
        if (!connected) {
            return "Not Connected: " + address;
        }

        if (name == null) {
            this.name = address;
        }
        return "JamochaMUD (" + name +")";
    }

    /** This returns the true name of the MU*, regardless of our connection
     */
    public String getMUName() {
        return this.name;
    }

    /** Call this to reset the frame's title to the &quot;standard" title.
     * This is usually just the location */
    public synchronized void resetTitle() {
        // this.name = "JamochaMUD (" + address + ")";
        this.name = name;
        setFrameTitles();
    }

    /** The Address of our connection */
    public void setAddress(String address) {
        this.address = address;
    }

    public String getAddress() {
        return address;
    }

    /** The port our connection will connect to */
    public void setPort(int port) {
        this.port = port;
    }

    public int getPort() {
        return port;
    }

    /** The timestamp of our connection, for internal use mostly */
    public void setTimeStamp(long time) {
        this.timeStamp = time;
    }

    public synchronized long getTimeStamp() {
        return timeStamp;
    }

    /** This is the main method for our thread.  Veryvery important!
     *  Basically, it'll be a big loop that only gets broken when our
     * connection disappears... either from the socket's end or the
     * user's end */
    public void run() {
        String tempAddy = this.getAddress();

        // Give some visual notification that we're attempting a connection
        setVisuals(tempAddy);

        // Write the location to the main window's title, as well

        // Try to establish a connection
        int port = this.getPort();
        InetAddress serverAddy;

        // We'll give sockPort and sockHost initial values to satisfy later conditions
        int sockPort = 0;
        String sockHost = "";
        boolean socks = settings.getProxy();

        if (socks) {
            sockHost = settings.getProxyHost();
            sockPort = settings.getProxyPort();
        }

        try {
            serverAddy = InetAddress.getByName(tempAddy);
            if (socks) {
                System.out.println("Attempt to use Socks5socket.");
                connSock = (Socket)new Socks5socket(tempAddy, port, sockHost, sockPort);
                System.out.println("Socks5socket successfully connected");
            } else {
                connSock = new Socket(serverAddy, port);
            }
            outStream = new DataOutputStream(connSock.getOutputStream());
        } catch (NoRouteToHostException oops) {
            // No route to host, or operation timed out
            disconnectMenu();

            String tempString = new String(oops + "");
            if (tempString.endsWith("unreachable")) {
                // Host unreachable
                mainText.append(RB("noRouteToHostException") + '\n');
            } else {
                mainText.append(RB("operationTimedOutException") + '\n');
            }
        } catch (UnknownHostException oops) {
            disconnectMenu();
            mainText.append(RB("unknownHostException") + '\n');
        } catch (ConnectException oops) {
            disconnectMenu();
            mainText.append(RB("connectException") + '\n');
        } catch (Exception oops) {
            disconnectMenu();
            System.out.println("From Net, no socket " + oops);

            mainText.append(RB("exception") + '\n');
            // Fix this XXX
            // closeConnection();
        }

        // Create a MUDBufferedReader for our input stream from the MU*
        InputStreamReader inStreamReader;

        try {
            inStreamReader = new InputStreamReader(connSock.getInputStream());
        } catch (Exception e) {
            // System.out.println("Error creating InputStreamReader.");
            return;
        }

        try {
            inStream = new MUDBufferedReader(inStreamReader);
        } catch (Exception e) {
            // System.out.println("No inStream");
            mainText.append("---> " + e + " <---" + '\n');
            closeCleanUp();
            return;
        }

        inStream.setEOR(true);  // We want to have EOR's if we can

        /*  Why the heck is this block in here twice?
         *  Thanks to Leonid for pointing this out.
         // Try and setup a MUDBufferedReader for our "data-pipe"
         try {
         inStream = new MUDBufferedReader(new InputStreamReader(connSock.getInputStream()));
         } catch (IOException oops) {
         // System.out.println("No inStream");
         mainText.append("---> " + oops + " <---" + '\n');
         closeCleanUp();
         return;
         }
         */

        // Okay, we're connected.  Do our little set-up stuff

        // Some specialised MU*'s require a form of authentication immediately
        // after connection.  This could be done here by adding something like:
        // sendText("foo");

        // Read from the socket until we get tired and fall down
        connected = true; // This notifies the program of true connection

        // Change the titles of our frames to show we're connected
        setFrameTitles();

        // Update the main window menus to reflect our connected state
        connectMenu();

        while (connected) {
            try {
                fromMU = inStream.readLine();
            } catch (Exception e) {
                // We've run into some sort've problem, so chances are
                // our connection has been terminated in some way.
                // System.out.println("Error at FromNet inStream.readLine " + e);
                // e.printStackTrace();
                // closeConnection();
                connected = false;
                // break;
                fromMU = null;
            }
            // Check for output plugins
            // Fix this XXX
            // we should probably not be making a "static" call like this

            // Check to see if we somehow got disconnected.  (It can happen!)
            if (fromMU == null) {
                // sets out menu to the correct state as well as closes the socket
                mainText.append("---> " + RB("connectionClosed") + " <---" + '\n');
                closeSocket();
                connected = false;
                disconnectMenu();
                return;
            }

            fromMU = EnumPlugIns.CallPlugin(fromMU, "output", this);

            // Send the current line to our Text-field
            if (!paused) {
                // Check for any IAC messages and parse them before we go forward
                // if (fromMU.indexOf(IAC) > 0) {
                while (fromMU.indexOf(IAC) > -1) {
                    // System.out.println("contains IAC at position " + fromMU.indexOf(IAC));
                    fromMU = checkIAC(fromMU);
                }

                // Check for EOR markings
                // while (fromMU.indexOf(EOR) > -1) {
                // System.out.println("Hit EOR");
                // System.exit(0);
                // }

                // Check for any URLs
                fromMU = checkURL(fromMU);

                mainText.append(fromMU);

                // for (int i = 0; i < fromMU.length(); i++) {
                // System.out.print(fromMU.charAt(i));
                // System.out.print("(" + Integer.toString((int)fromMU.charAt(i) , 16 /* radix */ ) + ")");
                // if (Character.isISOControl(fromMU.charAt(i))) {
                // System.out.print(Character.getNumericValue(fromMU.charAt(i)));
                // }
                // }
                // System.out.println("");

                // We'll notify the main window if needed
                if (!isActiveMU()) {
                    if (activity < 1) {
                        // Append an activity notice to the active window
                        connHandler.getActiveMUDText().append(RB("activityOn") + " " + getMUName() + '\n');
                        activity++;
                        // System.out.println("Add activity.");
                    }
                } else {
                    activity = 0;
                    // System.out.println("Reset activity.");
                }

                // Check to see if the main frame is minimised.  If so, we add
                // the new text to the title bar
                if (settings.isMainWindowIconified()) {
                    addTitleText(fromMU);
                } else {
                    minimisedLineCount = 0;
                }

            } else {
                // System.out.println("paused: " + fromMU);
                heldResponse.addElement(fromMU);
            }

        }

    }

    private static String RB(String itemTarget) {
        ResReader reader = new ResReader();
        // return reader.LangString("JamochaMUDBundle", itemTarget);
        return reader.LangString(JMConfig.BUNDLEBASE, itemTarget);
    }

    // Return our text area to who-ever wants to know! */
    public synchronized JMText getTextWindow() {
        return mainText;
    }

    /** Send text out to our MU* */
    public void sendText(String send) {
        try {
            if (settings.getUseUnicode()) {
                // Send the string as Unicode
                outStream.writeChars(send + "\r\n");
            } else {
                // Send the string as regular ANSI text
                outStream.writeBytes(send + "\r\n");
            }
        } catch (Exception e) {
            // Most likely we aren't connected to the MU*
            // System.out.println("MuSocket.sendText exception: " + e);
        }
    }

    /** Set the proper colours, fonts, etc */
    public void setMainColours() {
        mainText.setForeground(settings.getForegroundColour());
        mainText.setBackground(settings.getBackgroundColour());
        mainText.setFont(settings.getFontFace());
    }

    /** This will set up some of the initial indicators
     * that we're trying to make a connection */
    private void setVisuals(String title) {
        // mainText.append("Attempting connection to " + title + "..." + '\n');
        mainText.append(RB("attemptingConnectionTo") + " " + title + "..." + '\n');
    }

    private void setFrameTitles() {
        // We should only be able to change the frame titles it this
        // is the active MU*
        // CHandler connHandler = settings.getConnectionHandler();
        if (connHandler.getActiveMUHandle() == this) {
            MuckMain target = settings.getMainWindowVariable();
            target.setWindowTitle();

            DataIn inWindow = settings.getDataInVariable();
            inWindow.setWindowTitle();
        }
    }

    /** Return the status of our connection:<br>
     * <pre>true</pre> - Connected<br>
     * <pre>false</pre> - Not connected
     */
    public boolean isConnectionActive() {
        return connected;
    }

    /** Set the proper states of our windows, etc.
     * 'cause, woe is us, we are disconnected.  Most often
     * this would be called internally by our class
     */
    private void disconnectMenu() {
        // This is where we actually disconnect the socket.
        // Is there a nicer way of saying good-bye?
        // closeSocket();
        // closeSocket() is called from the CHandler class.
        // CHandler connHandler = settings.getConnectionHandler();

        // Do we really need to close the socket?!  We should get here because the
        // socket is already closed!  Check this out! XXX
        // This guarantees that we've cleaned up after ourselves, too
        // connHandler.closeSocket(this);

        // Change the connection notice on the main window if applicable
        MuckMain window = settings.getMainWindowVariable();
        window.CheckDisconnectMenu();

    }

    /** Allow other classes to close our socket */
    public void closeSocket() {
        if (DEBUG) {
            StringWriter sw = new StringWriter();
            new Throwable().printStackTrace(
                                            new PrintWriter( sw )
                                           );
            String callStack = sw.toString();

            System.out.println("Stack: " + callStack);
        }

        try {
            connSock.close();
        } catch (Exception closeError) {
            // The only reason we should get here is if the socket was never opened
            // in the first place.  This can happen if we tried to open a bad address
            // System.out.println("MuSocket hit an error (.disconnectMenu): " + closeError);
        }

        // Change the titles of the frames if applicable
        // System.out.println(getTitle() + " -> closed socket.");
        resetTitle();
        // setFrameTitles();

        return;
    }

    /** Set the proper states of our windows as
     * we've probably just connected
     */
    private void connectMenu() {
        MuckMain main = settings.getMainWindowVariable();
        main.setConnectionMenu();
    }

    /** Set the paused status of our text area */
    public void setPaused(boolean option) {
        paused = option;

        if (option) {
            setFrameTitles();
        } else {
            SpoolText();
        }
    }

    /** Return the fact that our connection is either
     * paused - <pre>true</pre><br>
     * or not paused - <pre>false</pre><br>
     */
    public synchronized boolean isPaused() {
        return paused;
    }

    public synchronized void SpoolText() {
        // We'll reset this variable, just in case
        paused = false;

        // System.out.println("Spooltext, " + heldResponse.size() + " entries.");

        // Reset our frame titles
        resetTitle();

        // Do a little bounds checking first
        if (heldResponse.size() < 1) {
            /* System.out.println("MuSocket: SpoolText, nothing to spool."); */
            return;
        }

        // System.out.println("MuSocket.SpoolText: we have " + heldResponse.size() + " elements.");
        while (heldResponse.size() > 0) {
            // System.out.println("MuSocket.SpoolText: " + heldResponse.elementAt(0).toString());
            mainText.append(heldResponse.elementAt(0).toString());
            System.out.println(heldResponse.elementAt(0).toString());
            heldResponse.removeElementAt(0);
        }
    }

    /** Trim down the string we've received from the MU*, add
     * the minimised line count, and the set the new title */
    private synchronized void addTitleText(String newText) {
        minimisedLineCount++;
        // System.out.println("MuSocket.addTitleText: " + minimisedLineCount + " " + newText);

        String str = StripEscapes(newText);

        // Trim the str down to a managable length
        String trimmedString;
        if (str.length() > 80) {
            trimmedString = str.substring(0, 80) + "...";
        } else {
            trimmedString = str;
        }

        MuckMain target = settings.getMainWindowVariable();
        target.setWindowTitle("(" + minimisedLineCount + ") " + str);

    }

    /** This method returns the value of a string
     * once it has had any escape characters stripped
     * from it.
     */
    private String StripEscapes(String token) {
        StringBuffer workString = new StringBuffer("");
        boolean loop = true;
        int start = 0;
        int end = 0;

        while (loop) {
            end = token.indexOf('\u001b', start);

            if (end < start) {
                // There are no escapes left
                workString.append(token.substring(start));
                loop = false;
                break;
            }

            if (end > 0) {
                workString.append(token.substring(start, end));
            }

            // Now set the new 'start'
            start = token.indexOf('m', end) + 1;
            if (start <= end) {
                loop =false;
                System.out.println("Not a proper ANSI escape");
                break;
            }
        }

        return workString.toString();
    }

    private void closeCleanUp() {
        closeSocket();
        connected = false;
        disconnectMenu();
    }

    /** Returns <pre>true</pre> if this connection is the currently active
     * (visible to the user) MU* connection.  <pre>False</pre> is returned
     * otherwise.
     */
    private boolean isActiveMU() {
        return activeStatus;
    }

    /** Indicate that this MU active or inactive */
    public void setActiveMU(boolean status) {
        activeStatus = status;
        // System.out.println("active state: " + status + " for " + getTitle());

        // If this MU* is set active then we can reset our activity counter
        if (status) {
            activity = 0;
        }
    }

    /** Check IAC codes and do our necessary magic */
    private String checkIAC(String input) {
        // System.out.println("Original: " + input);
        StringBuffer output = new StringBuffer("");
        int trimspot = 0, trimlen = 0;
        boolean known = false;  // Whether we know what this IAC is or not

        // Cheap and ugly echo stuff
        if (input.indexOf(IAC_WONT_ECHO) > -1) {
            trimspot = input.indexOf(IAC_WONT_ECHO);
            trimlen = IAC_WONT_ECHO.length();
            // trimlen = 3;
            echo = false;
        }

        if (input.indexOf(IAC_WILL_ECHO) > -1) {
            trimspot = input.indexOf(IAC_WILL_ECHO);
            trimlen = IAC_WILL_ECHO.length();
            echo = true;
        }

        if (input.indexOf(IAC_WILL_EOR) > -1) {
            // System.out.println("Sending IAC_DO_EOR");
            // Do we need an END-OF-RECORD?  Sure!!
            trimspot = input.indexOf(IAC_WILL_EOR);
            trimlen = IAC_WILL_EOR.length();

            // It looks like we will have to return this as soon as we receive it
            // or the MU* will not know to expect it back and will cough up an error
            // sendText(IAC_DO_EOR);
        }

        // Nothing changes the "known" variable yet!  Fix me XXX
        // if (!known) {
        // We don't know what type of IAC this is, but should snip it
        // so we don't end up in an endless loop
        // System.out.println("Uknown IAC... trimming.");
        trimspot = input.indexOf(IAC);
        trimlen = 2;
        // }

        if (trimspot > 0) {
            output.append(input.substring(0, trimspot));
            // System.out.println("Trim 1: " + output.toString());
        }

        if (trimspot + trimlen + 1 < input.length()) {
            output.append(input.substring(trimspot + trimlen + 1, input.length()));
            // System.out.println("Trim 2: " + output.toString());
        }

        return output.toString();
    }

    /** Return the state of our "echo" */
    public boolean getEchoState() {
        return echo;
    }

    /** Check the string for anything understood as a URL and underline it */
    private String checkURL(String input) {
        String checkChunk, lowerCheckChunk;
        StringBuffer in = new StringBuffer(input);
        int chunkSize = 0;

        java.util.StringTokenizer tokens = new java.util.StringTokenizer(input);

        while (tokens.hasMoreTokens()) {
            checkChunk = tokens.nextToken();
            lowerCheckChunk = checkChunk.toLowerCase();
            if (lowerCheckChunk.indexOf("www.") >= 0 || lowerCheckChunk.indexOf("http://") >= 0) {
                int match;
                match = (in.toString()).indexOf(checkChunk);

                in.insert(match + checkChunk.length(), "\u001b[24m");
                in.insert(match, "\u001b[4m");

            }
        }

        return in.toString();
    }
}
