/** MUDBufferedReader, a modified BufferedReader for JamochaMUD,
 * Portions rewritten by Jeff Robinson 2001/07/15
 * most code in this class is directly from the java.io.BufferedReader
 * that ships with Kaffe, by Transvirtual Technologies, Inc.  (See notice below)
 */

/*
 * Java core library component.
 *
 * Copyright (c) 1997, 1998
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 */

// package java.io;
package anecho.extranet;

import java.io.DataOutputStream;
import java.io.Reader;
import java.io.IOException;

public class MUDBufferedReader extends Reader {
    private static final int DEFAULTBUFFERSIZE = 8192;
    private Reader rd;
    DataOutputStream out;
    private char[] inbuf;
    private int pos;	// position of next char in buffer
    private int size;	// total length of valid chars in buffer
    //  invariant: 0 <= pos <= size <= inbuf.length
    private boolean markset;
    private boolean markvalid;
    private boolean setEORMark = false; // Do we receive EOR marks from the MU*?
    private boolean isIAC = false;
    private boolean isWILL = false;
    private boolean isEOR = false;

    // Keep track of any programmatic choices of IAC responses
    private boolean MBR_EOR = true;     // Should be set "false" by default XXX

    private long prompt_sec = 0; // Number of milliseconds we've waited since the last character
    private static int prompt_usec = 250; // Maximum wait-time for a new character
    // Is prompt_usec still needed with our new reading format? XXX
    // private static int prompt_usec = 2500; // Maximum wait-time for a new character

    private static final char IAC = '\u00a0';   // IAC
    private static final char EOR = '\u0019';       // EOR End of record (line)
    private static final char WILL = '\u00b9';        // WILL
    private static final char DO = '\u00fd';    // DO
    private static final char DEAD = '\uffff';  // Disconnected MU*?
    private static final char NULL = '\u0000';  // NULL
    private static final String CR = "\r";    // Carraige Return
    // private static final String EOL = System.getProperty("line.separator");
    private static final String IAC_DO_EOR = new String("\u00a0" + "\u00fd" + "\u0019");



    public MUDBufferedReader(Reader in) {
        this(in, DEFAULTBUFFERSIZE);
    }

    public MUDBufferedReader(Reader in, DataOutputStream out) {
        this(in, out, DEFAULTBUFFERSIZE);
    }

    public MUDBufferedReader(Reader in, int sz) {
        this (in, null, sz);
    }

    public MUDBufferedReader(Reader in, DataOutputStream out, int sz) {
        super(in);
        this.out = out;

        if (sz <= 0) {
            throw new IllegalArgumentException("Buffer size <= 0");
        }
        rd = in;
        inbuf = new char[sz];
    }


    /* Internal function used to check whether the
     BufferedReader has been closed already, throws
     an IOException in that case.
     */
    private void checkIfStillOpen() throws IOException {
        if (rd == null) {
            throw new IOException("Stream closed");
        }
    }

    /** Close our *quot;socket" and reset our key variables
     */
    public void close() throws IOException {
        synchronized(lock) {
            if (rd == null) {
                return;
            }

            // Close the input reader
            rd.close();

            // Release the buffer
            inbuf = null;

            // Release the input reader (lock is a reference too)
            rd = null;
            lock = this;
        }
    }

    public void mark(int readAheadLimit) throws IOException {
        if (readAheadLimit < 0) {
            throw (new IllegalArgumentException("Read-ahead limit < 0"));
        }

        synchronized(lock) {
            checkIfStillOpen();

            char[] oldbuf = inbuf;

            // Allocate bigger buffer if necessary
            if (readAheadLimit > inbuf.length) {
                inbuf = new char[readAheadLimit];
            }

            // Shift data to the beginning of the buffer
            System.arraycopy(oldbuf, pos, inbuf, 0, size - pos);
            size -= pos;
            pos = 0;
            markset = true;
            markvalid = true;
        }
    }

    public boolean markSupported() {
        return true;
    }

    //Used in java.io.BufferedLineReader
    void pushback() {
        synchronized (lock) {
            pos--;
        }
    }

    public int read () throws IOException {
        synchronized (lock) {
            checkIfStillOpen();

            if (pos < size || fillOutBuffer() > 0) {
                return (inbuf[pos++]);
            }
        }
        return (-1);
    }

    public int read ( char [] cbuf, int off, int len ) throws IOException {
        if (off < 0 || off + len > cbuf.length || len < 0) {
            throw new IndexOutOfBoundsException();
        }
        synchronized(lock) {
            checkIfStillOpen();

            // Do read directly in this case, according to JDK 1.2 docs
            if (pos == size && !markset && len >= inbuf.length) {
                // return rd.read(cbuf, off, len);
                return readCheck(cbuf, off, len);
            }

            int nread;
            int chunk;
            for (nread = 0; nread < len; nread += chunk) {
                // Make sure there's something in the buffer
                if (pos == size) {
                    // Avoid unneccessary blocking
                    if (nread > 0 && !rd.ready()) {
                        return nread;
                    }
                    // Try to fill buffer
                    if (fillOutBuffer() <= 0) {
                        return (nread > 0) ? nread : -1;
                    }
                }

                // Copy next chunk of chars from buffer to output array
                chunk = len - nread;
                if (chunk > size - pos) {
                    chunk = size - pos;
                }
                System.arraycopy(inbuf, pos, cbuf, off, chunk);
                pos += chunk;
                off += chunk;
            }
            return (nread);
        }
    }

    public String readLine () throws IOException {
        char    c = ' ';    // silly javac, complains about initialization
        String  s = null;

        synchronized ( lock ) {
            checkIfStillOpen();

            int start = pos;

            while ( true ) {

                // Find next newline or carriage return
                // while (pos < size
                // && (c = inbuf[pos]) != '\n' && c != '\r' && c != EOR) {
                // pos++;
                // }
                while (pos < size
                       && !isNewLine(inbuf[pos])) {
                    pos++;
                }

                // Did we see one?
                if (pos == size) {      // nothing found yet

                    if (pos > start) {
                        if (s == null) {
                            s = new String(inbuf,
                                           start, pos-start);
                            // return (s + '\n');
                            // Here, we could send back our string if the timer
                            // runs out
                        } else {
                            s += new String(inbuf,
                                            start, pos-start);
                        }
                    }

                    // We want to be sure that there is still more input left
                    // before we call fillOutBuffer, otherwise, we 'hang'
                    // Temporarily try the new method XXX
                    // prompt_sec = System.currentTimeMillis();  // set our timer
                    // setSecTimer();

                    boolean streamReady = rd.ready();

                    if (s != null) {
                        while (!streamReady && s.length() > 0) {
                            // if (!setEORMark) {
                                if (System.currentTimeMillis() - prompt_sec >= prompt_usec) {

                                    // We don't want to return a totally empty string
                                    if (s.length() > 0) {
                                        // System.out.println("prompt_usec surpassed.");
                                        // return (s + '\n');
                                        return s;
                                        // return ("*" + s + "*");

                                    } else {
                                        System.out.println("Zero-length string");
                                    }
                                }
                                // }
                            streamReady = rd.ready();
                        }
                    }

                    if (fillOutBuffer() < 0) {
                        return (s);
                    }
                    start = 0;
                } else {			// we got a line terminator
                    if ( s == null ) {
                        s = new String(inbuf, start, pos-start);
                    }
                    else {
                        s += new String(inbuf,
                                        start, pos-start);
                    }
                    pos++;
                    if (c == '\r') {
                        char [] buf = new char [1];
                        int     n;
                        while ((n = read(buf, 0, 1)) == 0)
                            ;

                        if (n == 1
                            && buf[0] != '\n' && pos > 0) {
                            pos--;  // skip over "\r\n"
                        }
                    }

                    // System.out.println("CR: " + s.indexOf("\n"));
                    if (setEORMark || s.indexOf(CR) > -1) {
                        // System.out.println("No return.");
                        return s;
                    } else {
                        // System.out.println("With return.");
                        return (s + '\n');
                        // return s;
                    }

                }
            }
        }
    }

    public boolean ready() throws IOException {
        synchronized(lock) {
            checkIfStillOpen();

            return (pos < size || rd.ready());
        }
    }

    // This only gets called when pos == size. It fills as much of the buffer
    // beyond position "size" as possible. If no mark is set, we shift "pos"
    // back to zero; otherwise, don't shift the buffer unless the mark overflows.

    private int fillOutBuffer () throws IOException {
        synchronized ( lock ) {
            int n;

            if (markset) {
                if (pos == inbuf.length) {	// mark overflow
                    markvalid = false;
                    pos = 0;
                }
            } else {
                pos = 0;
            }

            // n = rd.read(inbuf, pos, inbuf.length - pos);
            n = readCheck(inbuf, pos, inbuf.length - pos);

            if (n >= 0) {
                size = pos + n;
                return (n);
            }
            else {
                size = pos;
                return (-1);
            }
        }
    }

    public void reset() throws IOException {
        synchronized(lock) {
            checkIfStillOpen();

            if (!markset) {
                throw new IOException("Stream not marked");
            }
            else if (!markvalid) {
                throw new IOException("Mark invalid");
            }
            pos = 0;
        }
    }

    public long skip(long n) throws IOException {
        if (n < 0) {
            throw new IllegalArgumentException("skip value is negative");
        }
        synchronized(lock) {
            checkIfStillOpen();

            long bufskip;

            // Skip from within the buffer first
            bufskip = size - pos;
            if (bufskip > n) {
                bufskip = n;
            }
            pos += (int)bufskip;            // cast is OK

            // Anything more to skip?
            if (bufskip == n) {
                return n;
            }

            // Reset buffer and skip what remains directly
            pos = 0;
            size = 0;
            markvalid = false;
            return bufskip + rd.skip(n - bufskip);
        }
    }

    /** Checks to see if the current character passed is a "new-line"
     * character or not.
     * @return  true
     * @return  false
     */
    private boolean isNewLine(char c) {
        boolean result = false;
        // if (c == '\n' || c == '\r' || c == EOR) {
        if (c == '\n' || c == '\r') {
            result = true;
        }

        return result;

    }

    /** Read in more from our reader, but check for any AIC commands
     */
    private synchronized int readCheck(char[] cbuf, int off, int len) throws IOException {
    // private int readCheck(char[] cbuf, int off, int len) {
        // System.out.print("rc");
        // boolean isIAC = false;
        // boolean isWILL = false;
        // boolean isEOR = false;

        int end = (off + len);
        int inChar = 0;
        char testChar;
        int bufLength = 0;

        for (int i = off; i < end; i++) {
            inChar = rd.read();
            testChar = (char)inChar;

            if (testChar == NULL) {
                System.out.print("Blank ");
            }

            if (inChar < 0 || testChar == DEAD) {
                // System.out.println("Negative inChar.");
                // }
                // if (testChar == DEAD) {
                close();
            }

            // Check for our IACs
            switch (testChar) {
            case IAC:
                isIAC = true;
                // System.out.println("IAC");
                break;

            case WILL:
                isWILL = true;
                // System.out.println("WILL");
                break;

            case EOR:
                if (isIAC && isWILL) {
                    // System.out.println("IAC WILL EOR!!!!!!!");
                    // System.out.println("IWE");
                    isIAC = false;
                    isWILL = false;
                    isEOR = false;

                    if (willEOR() && out != null) {
                        // Send a response back that we will EOR!
                        // outStream.writeChars(send + "\r\n");
                        out.writeChars(IAC_DO_EOR);
                        System.out.println("Sent IAC_DO_EOR");
                    }
                    // bufLength = bufLength - 3;
                } else {
                    isEOR = true;
                }
            }

            if (testChar == '\r') {
                // Right now we have to replace the Carraige Return with a space as
                // we cannot seem to successfully detect it once we have a string for output!?
                testChar = ' ';
            }

            cbuf[bufLength] = testChar;
            bufLength++;

            if (!rd.ready()) {
                // System.out.println(" (" + bufLength +") ");
                // Let's try setting our timer here out of interest
                setSecTimer();

                return bufLength;
            }

            // System.out.print(testChar + "("+inChar+") ");
            // System.out.print(testChar);

        }

        // return rd.read(cbuf, off, len);
        // System.out.println("bufLength: " + bufLength);
        // System.out.println(" ebl(" + bufLength + ")");

        setSecTimer();

        return bufLength;

    }

    private void setSecTimer() {
        prompt_sec = System.currentTimeMillis();  // set our timer
    }

    /** Should we respond to a request for IAC EOR? */
    public void setEOR(boolean state) {
        if (out != null) {
            MBR_EOR = state;
        }
    }

    /** Our IAC EOR response status */
    public boolean willEOR() {
        return MBR_EOR;
    }

}
